
(* *********************************** *)
(* Definition of the abstract machine  *)
(* *********************************** *)


Require Import Map.
Require Import EqNat.
Require Import Arith.

Definition Word := nat.

Definition Address := Word.

(* eq_nat_dec is defined in Coq library Coq.Arith.Peano_dec *)
Definition beq_w := eq_nat_dec.

Inductive Lock : Set :=
  | l1 | l2.


Definition beq_l : forall (l l' : Lock), {l = l'} + {l <> l'}.
Proof.
  double induction l l'.
  left. auto. right. discriminate.
  right. discriminate. left. auto.
Defined.


(*
Definition beq_l (l l' : Lock) :=
  match l,l' with
  | l1,l1 => true
  | l2,l2 => true
  | _,_ => false
  end.
*)

Inductive Tid : Set :=
  | t1 | t2.


Definition beq_t : forall (t t' : Tid), {t = t'} + {t <> t'}.
Proof.
  double induction t t'.
  left. auto. right. discriminate.
  right. discriminate. left. auto.
Defined.


(*
Definition beq_t (t t' : Tid) :=
  match t,t' with
  | t1,t1 => true
  | t2,t2 => true
  | _,_ => false
  end.
*)

Inductive  Reg : Set :=
  | r0 | r1 | r2 | r3 | r4 | r5 | r6 | r7.

Definition beq_r (r r' : Reg) : bool :=
  match r,r' with
  | r0,r0 => true
  | r1,r1 => true
  | r2,r2 => true
  | r3,r3 => true
  | r4,r4 => true
  | r5,r5 => true
  | r6,r6 => true
  | r7,r7 => true
  | _,_ => false
  end.

Definition RegFile := Reg -> Word.

Definition ext_R (R : RegFile) (r : Reg) (w : Word) :=
  fun r' => if beq_r r' r
            then w
            else R r'.


Definition Mem := Map Address Word.

Definition has_M (M : Mem) (l : Address) : Prop := in_dom l M.

Definition ext_M (M: Mem) := put beq_w M.

Definition get_M (M: Mem) := maps_to M.

Definition uni_M (M1 M2 : Mem) : Mem := merge M1 M2.

Definition dis_M (M1 M2 : Mem) : Prop := disjoint M1 M2.

Definition remove_M (M : Mem) := remove beq_w M.

Definition sub_M (M M' : Mem) : Prop := subseteq M M'.

Inductive Instr : Set :=
  | addu (rd rs rt : Reg)
  | cons (rd : Reg)
  | free (rs : Reg)
  | sw (rt : Reg) (w : Word) (rs : Reg)
  | lock (l : Lock)
  | unlock (l : Lock)
  | bgtz (rs : Reg) (f : Address).

Inductive InstrSeq : Set :=
  | seq (c : Instr) (I : InstrSeq)
  | jr (rs : Reg).


Definition State := (Mem * RegFile)%type.

Definition CodeHeap :=  Map Address InstrSeq.

Definition Thread := (CodeHeap * RegFile * InstrSeq * Tid)%type.

Definition LockMap := Map Lock Tid.


Definition ins_L (L : LockMap) (l : Lock) (k : Tid) : LockMap := put beq_l L l k.


Definition del_L (L : LockMap) (l : Lock) : LockMap := remove beq_l L l.


Definition filter_L (L : LockMap) (t : Tid) : LockMap :=
  fun l => match L l with
	   | Some t' => if beq_t t t'
	                then Some t
			else None
	   | None => None
	   end.


Definition Program := (Mem * (Thread * Thread) * LockMap)%type.

Definition XState := (Mem * (RegFile * Tid) * LockMap)%type.

Inductive Next : Instr -> State -> State -> Prop :=
  | Next_addu : forall M R rd rs rt, Next (addu rd rs rt) (M,R) (M, ext_R R rd (R rs+R rt))
  | Next_sw : forall M R rt w rs, has_M M (R rs+w) -> Next (sw rt w rs) (M,R) (ext_M M (R rs+w) (R rt), R)
  | Next_cons : forall M R rd l w0 w1,
		    not_in_dom l M ->
		    not_in_dom (S l) M ->
		    Next (cons rd) (M, R) (ext_M (ext_M M l w0) (S l) w1, ext_R R rd l)
  | Next_free : forall M R rs,
                    in_dom (R rs) M ->
		    Next (free rs) (M, R) (remove_M M (R rs), R).

Inductive ThrdStep : (Mem * Thread * LockMap) -> (Mem * Thread * LockMap) -> Prop :=
  | ThrdStep_c : 
        forall M C R c I k L M' R', 
	  Next c (M,R) (M',R') -> ThrdStep (M,(C,R,seq c I,k),L) (M',(C,R',I,k),L)
  | ThrdStep_jr : 
        forall M C R rs k L I', 
	  maps_to C (R rs) I' -> ThrdStep (M,(C,R,jr rs,k),L) (M,(C,R,I',k),L)
  | ThrdStep_lock : 
        forall M C R l I k L, 
	  not_in_dom l L -> ThrdStep (M,(C,R,seq (lock l) I,k),L) (M,(C,R,I,k),ins_L L l k)
  | ThrdStep_lockwait : 
        forall M C R l I k L, 
	  in_dom l L  -> ThrdStep (M,(C,R,seq (lock l) I,k),L) (M,(C,R,seq (lock l) I,k),L)
  | ThrdStep_unlock : 
        forall M C R l I k L, 
	  maps_to L l k -> ThrdStep (M,(C,R,seq (unlock l) I,k),L) (M,(C,R,I,k),del_L L l)
  | ThrdStep_bgtz_a :
        forall M C R k L rs f I I',
	  R rs > 0 -> maps_to C f I' -> ThrdStep (M, (C, R, seq (bgtz rs f) I, k), L)
	                                         (M, (C, R, I', k), L)
  | ThrdStep_bgtz_b :
        forall M C R k L rs f I,
	  R rs <= 0 -> ThrdStep (M, (C, R, seq (bgtz rs f) I, k), L) (M, (C, R, I, k), L).


Inductive Step : Program -> Program -> Prop :=
  | Step_t1 : 
        forall M M' T1 T1' T2 L L',
              ThrdStep (M, T1, L) (M', T1', L') ->
              Step (M, (T1, T2), L) (M',(T1', T2),L')
  | Step_t2 : 
        forall M M' T1 T2 T2' L L',
              ThrdStep (M, T2, L) (M', T2', L') ->
              Step (M,(T1, T2),L) (M',(T1, T2'),L').




(* ******************************** *)
(* Some properties of mem operation *)
(* More properties can be adapted   *)
(* from Map.v                       *)
(* ******************************** *)

Lemma has_uni_intro : forall M M' w, has_M M' w -> has_M (uni_M M M') w.
Proof.
  intros.
  unfold has_M in H.
  unfold uni_M. unfold has_M.
  apply merge_presv_in_dom_right; auto.
Qed.


Lemma get_sub_equal : forall M M' l w, sub_M M M' -> has_M M l -> get_M M l w = get_M M' l w.
Proof.
  unfold sub_M. unfold has_M. unfold get_M.
  unfold subseteq.
  intros.
  unfold maps_to.
  assert (M l = M' l). apply H. auto.
  rewrite H1. auto.
Qed.

Lemma uni_sub_impll : forall M M1 M2, M = uni_M M1 M2 -> sub_M M1 M.
Proof.
  intros.
  rewrite H.
  unfold uni_M.
  unfold sub_M.
  apply merge_extends_map.
Qed.

Lemma uni_sub_implr : forall M M1 M2, dis_M M1 M2 -> M = uni_M M1 M2 -> sub_M M2 M.
Proof.
  unfold dis_M.
  unfold uni_M.
  intros.
  assert (merge M1 M2 = merge M2 M1).
  apply disj_merge_commut. auto.
  rewrite H1 in H0.
  apply uni_sub_impll with M1; auto.
Qed.




